/*
    Copyright (C) 2001, 2002 Stephane Magnenat
    for any question or comment contact us at nct@ysagoon.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. | See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA | 02111-1307 | USA
*/

#include "SplineHighMap.h"
#include "SplineEditor.h"
#include <stdlib.h>
#include <assert.h>
#include <GL/gl.h>
#include <GL/glu.h>

SplineEditor::SplineEditor(float eyeX, float eyeY, float eyeZ, float screenSize, int res)
{
	if (res)
		this->triangleRes=res;
	else
		this->triangleRes=64;
	this->eyeX=eyeX;
	this->eyeY=eyeY;
	this->eyeZ=eyeZ;
	this->screenSize=screenSize;
	editX=0;
	editY=0;
	useStereo=false;
	canUseStereo=false;
	heightmap.Reset(7, 7);
	landscapeList=0;
	pointerList=0;
	rot=0.0f;
	useRotation=true;
}

SplineEditor::~SplineEditor()
{
	glDeleteLists(landscapeList, 1);
	glDeleteLists(pointerList, 1);
}

bool SplineEditor::init(void)
{
	if((SDL_Init(SDL_INIT_VIDEO)==-1))
	{
		fprintf(stderr, "Could not initialize SDL: %s.\n", SDL_GetError());
		return false;
	}

	atexit(SDL_Quit);

	return setVideoMode(640, 640);
}

bool SplineEditor::setVideoMode(int w, int h)
{
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 24 );
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
	screen = SDL_SetVideoMode(w, h, 32, SDL_ANYFORMAT|SDL_OPENGL|SDL_RESIZABLE);
	if ( screen != NULL )
	{
		glViewport(0, 0, screen->w, screen->h);
		prepareStencilBuffer();
		return true;
	}

	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
	screen = SDL_SetVideoMode(w, h, 16, SDL_ANYFORMAT|SDL_OPENGL|SDL_RESIZABLE);
	if ( screen != NULL )
	{
		glViewport(0, 0, screen->w, screen->h);
		prepareStencilBuffer();
		return true;
	}

	fprintf(stderr, "Couldn't set any %dx%d gl video mode !\n", w, h);
	return false;
}

void SplineEditor::prepareStencilBuffer(void)
{
	glEnable(GL_STENCIL_TEST);
	if (!glIsEnabled(GL_STENCIL_TEST))
	{
		fprintf(stderr, "Can't enable stencil buffer\n");
		canUseStereo=false;
		SDL_WM_SetCaption("Cologram, by Stephane Magnenat - Stereo disabled (Stencil buffer unavailable)", "cologram");
	}
	else
	{
		canUseStereo=true;
		SDL_WM_SetCaption("Cologram, by Stephane Magnenat - Stereo Enabled (Stencil buffer available)", "cologram");
	}
	glClearStencil(0x0);
	glClear(GL_STENCIL_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, screen->w, 0, screen->h);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glStencilFunc (GL_ALWAYS, 0x1, 0x1);
	glStencilOp (GL_REPLACE, GL_REPLACE, GL_REPLACE);
	glBegin(GL_LINES);
	for (int i=0; i<screen->w; i+=2)
	{
		glVertex2i(i, 0);
		glVertex2i(i, screen->h);
	}
	glEnd();
	if (!landscapeList)
		landscapeList=glGenLists(1);
	if (!pointerList)
		pointerList=glGenLists(1);
}

void SplineEditor::setProjectionMatrix(bool useStereo, bool side)
{
	float nearP=eyeZ-4;
	float farP=eyeZ+12;
	float matProjection[16];

	/*matProjection[0]=eyeZ/(screenSize/2);
	matProjection[1]=0.0f;
	matProjection[2]=(side ? 1 : -1)*eyeX/(screenSize/2);
	matProjection[3]=-(side ? 1 : -1)*eyeX*eyeZ/(screenSize/2);
	matProjection[4]=0.0f;
	matProjection[5]=eyeZ/(screenSize/2);
	matProjection[6]=eyeY/(screenSize/2);
	matProjection[7]=eyeY*eyeZ/(screenSize/2);
	matProjection[8]=0.0f;
	matProjection[9]=0.0f;
	matProjection[10]=farP/(farP-nearP);
	matProjection[11]=-nearP*farP/(farP-nearP); //-eyeZ;
	matProjection[12]=0.0f;
	matProjection[13]=0.0f;
	matProjection[14]=1.0f;
	matProjection[15]=0.0f;*/

	matProjection[0]=eyeZ/(screenSize/2);
	matProjection[4]=0.0f;
	if (useStereo)
	{
		matProjection[8]=(side ? 1 : -1)*eyeX/(screenSize/2);
		matProjection[12]=-(side ? 1 : -1)*eyeX*eyeZ/(screenSize/2);
	}
	else
	{
		matProjection[8]=0;
		matProjection[12]=0;
	}
	matProjection[1]=0.0f;
	matProjection[5]=eyeZ/(screenSize/2);
	matProjection[9]=eyeY/(screenSize/2);
	matProjection[13]=eyeY*eyeZ/(screenSize/2);
	matProjection[2]=0.0f;
	matProjection[6]=0.0f;
	matProjection[10]=farP/(farP-nearP);
	matProjection[14]=-nearP*farP/(farP-nearP); //-eyeZ;
	matProjection[3]=0.0f;
	matProjection[7]=0.0f;
	matProjection[11]=1.0f;
	matProjection[15]=0.0f;

	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(matProjection);
}

void SplineEditor::setMask(bool useStereo, bool side)
{
	if (useStereo)
	{
		if (side)
		{
			glStencilFunc (GL_EQUAL, 0x1, 0x1);
			glStencilOp (GL_KEEP, GL_KEEP, GL_KEEP);
		}
		else
		{
			glStencilFunc (GL_NOTEQUAL, 0x1, 0x1);
			glStencilOp (GL_KEEP, GL_KEEP, GL_KEEP);
		}
	}
	else
	{
		glStencilFunc(GL_ALWAYS, 0x0, 0x0);
		glStencilOp (GL_KEEP, GL_KEEP, GL_KEEP);
	}
}

bool SplineEditor::setTexture(const char *filename)
{
	SDL_Surface *surface;
	surface=SDL_LoadBMP(filename);//IMG_Load(filename);

	if (surface!=NULL)
	{
		glBindTexture(GL_TEXTURE_2D, texture);

		switch (surface->format->BitsPerPixel)
		{
			case 24:
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, surface->w, surface->h,  0,  GL_BGR, GL_UNSIGNED_BYTE, surface->pixels);
			break;

			case 32:
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, surface->w, surface->h,  0,  GL_BGRA, GL_UNSIGNED_BYTE, surface->pixels);
			break;

			default:
			SDL_FreeSurface(surface);
			return false;
		}

		// default parameters, to change
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

		SDL_FreeSurface(surface);

		return true;
	}
	return false;
}

void SplineEditor::renderTetrahedre(void)
{
    float points[4][3];
    float normals[4][3];
    int i;
    float rot, rotn;

#define M_PI 3.1415926535
#define DEG2RAD (3.1415926535/180.0f)
#define DONORMAL(n) glNormal3f(normals[n][0], normals[n][1], normals[n][2]);
#define DOPOINT(n) glVertex3f(points[n][0], points[n][1], points[n][2]);

    // generate coordinates
    rot=0.0f;
    rotn=-(2.0f*(float)M_PI)/6.0f;
    for (i=0; i<3; i++)
    {
        normals[i][0]=cos(rotn)*cos(DEG2RAD*30.0f);
        normals[i][1]=sin(rotn)*cos(DEG2RAD*30.0f);
        normals[i][2]=sin(DEG2RAD*30.0f);
        points[i][0]=cos(rot);
        points[i][1]=sin(rot);
        points[i][2]=0.0f;
        rot-=(2.0f*(float)M_PI)/3.0f;
        rotn-=(2.0f*(float)M_PI)/3.0f;
    }
    normals[3][0]=0.0f;
    normals[3][1]=0.0f;
    normals[3][2]=-1.0f;
    points[3][0]=0.0f;
    points[3][1]=0.0f;
    points[3][2]=3.0f/2.0f;

    glBegin(GL_TRIANGLES);
        DONORMAL(0);
        DOPOINT(0);
        DOPOINT(3);
        DOPOINT(1);

        DONORMAL(1);
        DOPOINT(1);
        DOPOINT(3);
        DOPOINT(2);

        DONORMAL(2);
        DOPOINT(2);
        DOPOINT(3);
        DOPOINT(0);

        DONORMAL(3);
        DOPOINT(0);
        DOPOINT(1);
        DOPOINT(2);
    glEnd();
}

void SplineEditor::renderLight(void)
{
	float ambient[4] = { 0.3f, 0.3f, 0.3f, 1.0f };
	float diffuse[4] = { 1.0f, 1.0f, 1.0f, 1.0f };
	float direction[4] = { 1.0f, 0.0f, 1.0f, 0.0f };
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
	glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0f);
	glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.0f);
	glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.0f);
	glLightf(GL_LIGHT0, GL_SPOT_CUTOFF, 180.0f);
	glLightfv(GL_LIGHT0, GL_POSITION, direction);
}

void SplineEditor::setMaterial(float r, float g, float b)
{
	float ambient[4] = { 0.3f, 0.3f, 0.3f, 1.0f };
	float diffuse[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
	diffuse[0]=r;
	diffuse[1]=g;
	diffuse[2]=b;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
}

void SplineEditor::renderLandscape(void)
{
	int x, y;
	int px, py;
	float pfx, pfy;
	float nx, ny, nz;
	float mapToScreen;
	float actHeight;
	float mw=heightmap.GetMapW();
	float mh=heightmap.GetMapH();
	const int w=triangleRes;
	const int h=triangleRes;

	nz=4.0f;
	float mapToScreenW=mw/((float)w);
	float mapToScreenH=mh/((float)h);
	for (x=0; x<w-1; x++)
	{
		glBegin(GL_TRIANGLE_STRIP);
		for (y=0; y<h*2; y++)
		{
			px=(x+(y&1));
			py=((y>>1));

			pfx=((float)px)/((float)w);
			pfy=((float)py)/((float)h);

			glTexCoord2f(pfx, pfy);

			actHeight=heightmap.GetH(pfx*mw, pfy*mh);
			nx=heightmap.GetH((px-1)*mapToScreenW, py*mapToScreenH)-actHeight;
			nx-=heightmap.GetH((px+1)*mapToScreenW, py*mapToScreenH)-actHeight;
			ny=heightmap.GetH(px*mapToScreenW, (py-1)*mapToScreenH)-actHeight;
			ny-=heightmap.GetH(px*mapToScreenW, (py+1)*mapToScreenH)-actHeight;

			glNormal3f(nx, ny, nz);

			glVertex3f((float)px, (float)py, actHeight);
		}
		glEnd();
	}
}

void SplineEditor::generateLists(void)
{
	glNewList(landscapeList, GL_COMPILE);
	renderLandscape();
	glEndList();
	glNewList(pointerList, GL_COMPILE);
	renderTetrahedre();
	glEndList();
}

void SplineEditor::renderScene(void)
{
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0f, 0.0f, 34.0f);

	glRotatef(-115.0f, 1.0f, 0.0f, 0.0f);
	glRotatef(rot, 0.0f, 0.0f, 1.0f);
	//glRotatef(rot, 1.0f, 0.0f, 0.0f);
	//glColor3f(1.0f, 0.0f, 1.0f);

	// landscape rendering from spline
	/*int x, y;
	int px, py;
	float pfx, pfy;
	float nx, ny, nz;
	float mapToScreen;
	float actHeight;*/
	const int w=triangleRes;
	const int h=triangleRes;
	const float scaleFactor=0.043f*(256.0f/triangleRes);
	float mw=heightmap.GetMapW();
	float mh=heightmap.GetMapH();

	renderLight();
	setMaterial( 1.0f, 1.0f, 1.0f);
	glTranslatef(-((float)w/2.0f)*scaleFactor, -((float)h/2.0f)*scaleFactor, 0.0f);
	glScalef(scaleFactor, scaleFactor, scaleFactor);
	glEnable(GL_TEXTURE_2D);
	glColor3f(1.0f, 1.0f, 1.0f);
	setMaterial( 1.0f, 1.0f, 1.0f);
	glCallList(landscapeList);
	//renderLandscape();

	float editfx=((float)editX*w)/mw;
	float editfy=((float)editY*h)/mh;
	glTranslatef(editfx, editfy, heightmap.GetCP(editX, editY)+2.4f/scaleFactor);
	glDisable(GL_TEXTURE_2D);
	glRotatef(-2*rot, 0.0f, 0.0f, 1.0f);
	glRotatef(180.0f, 0.0f, 1.0f, 0.0f);
	//glColor3f(0.5f+0.5f*sin(rot), 1.0f, 0.5f+0.5f*sin(rot));
	glColor3f(1.0f, 0.0f, 0.0f);
	glScalef(1.0f/scaleFactor, 1.0f/scaleFactor, 1.5f/scaleFactor);

	setMaterial( 1.0f, 0.0f, 0.0f);
	//renderTetrahedre();
	glCallList(pointerList);

	if (useRotation)
	{
		rot+=0.1f;
		if (rot>360.0f)
			rot-=360.0f;
	}
}

void SplineEditor::screenshot(const char *filename)
{
	SDL_Surface *image;
	SDL_Surface *temp;
	int idx;
	image = SDL_CreateRGBSurface(SDL_SWSURFACE, screen->w, screen->h,
								24, 0x0000FF, 0x00FF00, 0xFF0000, 0);
	temp = SDL_CreateRGBSurface(SDL_SWSURFACE, screen->w, screen->h,
								24, 0x0000FF, 0x00FF00, 0xFF0000, 0);

	glReadPixels(0, 0,  screen->w, screen->h, GL_RGB,
				GL_UNSIGNED_BYTE, image->pixels);
	for (idx = 0; idx < screen->h; idx++)
	{
			memcpy(((char *)temp->pixels) + 3 *  screen->w * idx,
					((char *)image->pixels) + 3
					*  screen->w*(screen->h - idx),
					3* screen->w);
	}
	memcpy(image->pixels,temp->pixels, screen->w * screen->h * 3);
	SDL_SaveBMP(image, filename);
	SDL_FreeSurface(image);
	SDL_FreeSurface(temp);
}

void SplineEditor::saveBitmaps(void)
{
	setMask(false);

	setProjectionMatrix(true, true);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	renderScene();
	glFlush();
	screenshot("left.bmp");

	setProjectionMatrix(true, false);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	renderScene();
	glFlush();
	screenshot("right.bmp");
}

int SplineEditor::run(void)
{
	SDL_Event event;
	bool isRunning=true;
	if (!init())
		return 1;

	glEnable(GL_TEXTURE_2D);
	setTexture("texture.bmp");
	// DO :
	//glDepthRange(eyeZ-2, eyeZ+10);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	//glEnable(GL_CULL_FACE);
	
	generateLists();

	while (isRunning)
	{
		while (SDL_PollEvent(&event))
		{
			if (event.type==SDL_QUIT)
			{
				isRunning=false;
			}
			else if (event.type==SDL_KEYDOWN)
			{
				switch (event.key.keysym.sym)
				{
					case SDLK_UP:
					editY--;
					if (editY<0)
						editY=heightmap.GetMapH();
					break;

					case SDLK_DOWN:
					editY++;
					if (editY>=heightmap.GetMapH())
						editY=0;
					break;

					case SDLK_LEFT:
					editX--;
					if (editX<0)
						editX=heightmap.GetMapW();
					break;

					case SDLK_RIGHT:
					editX++;
					if (editX>=heightmap.GetMapW())
						editX=0;
					break;

					case SDLK_PLUS:
					case SDLK_KP_PLUS:
					heightmap.SetCP(editX, editY, heightmap.GetCP(editX, editY)+0.03f*triangleRes);
					generateLists();
					break;

					case SDLK_MINUS:
					case SDLK_KP_MINUS:
					heightmap.SetCP(editX, editY, heightmap.GetCP(editX, editY)-0.03f*triangleRes);
					generateLists();
					break;
					
					case SDLK_s:
					if (canUseStereo)
						useStereo=!useStereo;
					break;
					
					case SDLK_r:
					rot=0.0f;
					break;

					case SDLK_b:
					saveBitmaps();
					break;
					
					case SDLK_SPACE:
					useRotation=!useRotation;
					break;

					default:
					break;
				}
			}
			else if (event.type==SDL_VIDEORESIZE)
			{
				setVideoMode(event.resize.w, event.resize.h);
			}
		}
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
		if (useStereo)
		{
			setProjectionMatrix(true, true);
			setMask(true, true);
			renderScene();
			setProjectionMatrix(true, false);
			setMask(true, false);
			renderScene();
		}
		else
		{
			setProjectionMatrix(false);
			setMask(false);
			renderScene();
		}
		glFlush();
		SDL_GL_SwapBuffers();
	}

	return 0;
}

int main(int argc, char *argv[])
{
	int tr=0;
	if (argc==2)
		tr=atoi(argv[1]);
	SplineEditor s(4.0f, 0.0f, 30.0f, 17.0f, tr);
	return s.run();
}
