/*
    Copyright (C) 2001, 2002 Stephane Magnenat
    for any question or comment contact us at nct@ysagoon.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. | See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA | 02111-1307 | USA
*/

#include "SplineHighMap.h"


void SplineHeightMap::Reset(int maskw,int maskh)
{
	cleanup();
	mapwmask=maskw;
	maphmask=maskh;
	mapw=maskw+1;
	maph=maskh+1;
	highmap=new float[mapw*maph];
	memset(highmap,0,mapw*maph*sizeof(float));
}

float SplineHeightMap::GetH(float x, float y)
{
	int ix=(int)x;
	int iy=(int)y;
	float u=x-(float)ix;
	float v=y-(float)iy;
	if (ix<0)
	{
		ix--;
		u=1.0f+u;
	}
	if (iy<0)
	{
		iy--;
		v=1.0f+v;
	}
	return Evaluate(u,v,ix,iy);
}
/*
inline void SplineHeightMap::GetCP4(int x,int y, float &f0, float &f1, float &f2, float &f3)
{
	if ((x>0) && (x<mapw-4))
	{
		float *hmp;
		hmp=highmap+x+(y&maphmask)*mapw;
		f0=*hmp++;
		f1=*hmp++;
		f2=*hmp++;
		f3=*hmp;
	}
	else
	{
		f0=GetCP(x,y);
		f1=GetCP(x+1,y);
		f2=GetCP(x+2,y);
		f3=GetCP(x+3,y);
	}
}
*/
float SplineHeightMap::Evaluate(float u, float v, int x, int y)
{
	float a,b,c,d;
	float cx0,cx1,cx2,cx3;
	float cy0,cy1,cy2,cy3;
	float u2=u*u;
	float u3=u2*u;
	float v2=v*v;
	float v3=v2*v;
	
	// compute 4 new control points
	// by=y-1
	cx0=GetCP(x-1,y-1);
	cx1=GetCP(x,y-1);
	cx2=GetCP(x+1,y-1);
	cx3=GetCP(x+2,y-1);
	a=/*0.5f*/(cx3-3*cx2+3*cx1-cx0);
	b=/*0.5f*/(-cx3+4*cx2-5*cx1+2*cx0);
	c=/*0.5f*/(cx2-cx0);
	d=2.0f*cx1;
	cy0=u3*a+u2*b+u*c+d;
	// by=y
	cx0=GetCP(x-1,y);
	cx1=GetCP(x,y);
	cx2=GetCP(x+1,y);
	cx3=GetCP(x+2,y);
	a=/*0.5f*/(cx3-3*cx2+3*cx1-cx0);
	b=/*0.5f*/(-cx3+4*cx2-5*cx1+2*cx0);
	c=/*0.5f*/(cx2-cx0);
	d=2.0f*cx1;
	cy1=u3*a+u2*b+u*c+d;
	// by=y+1
	cx0=GetCP(x-1,y+1);
	cx1=GetCP(x,y+1);
	cx2=GetCP(x+1,y+1);
	cx3=GetCP(x+2,y+1);
	a=/*0.5f*/(cx3-3*cx2+3*cx1-cx0);
	b=/*0.5f*/(-cx3+4*cx2-5*cx1+2*cx0);
	c=/*0.5f*/(cx2-cx0);
	d=2.0f*cx1;
	cy2=u3*a+u2*b+u*c+d;
	// by=y+2
	cx0=GetCP(x-1,y+2);
	cx1=GetCP(x,y+2);
	cx2=GetCP(x+1,y+2);
	cx3=GetCP(x+2,y+2);
	a=/*0.5f*/(cx3-3*cx2+3*cx1-cx0);
	b=/*0.5f*/(-cx3+4*cx2-5*cx1+2*cx0);
	c=/*0.5f*/(cx2-cx0);
	d=2.0f*cx1;
	cy3=u3*a+u2*b+u*c+d;
	// compute y spline
	a=/*0.5f*/(cy3-3*cy2+3*cy1-cy0);
	b=/*0.5f*/(-cy3+4*cy2-5*cy1+2*cy0);
	c=/*0.5f*/(cy2-cy0);
	d=2.0f*cy1;
	// compute final value
	return (a*v3+b*v2+c*v+d)*0.25f;
}